/* -----------------------------------------
  configparser.cpp
  
  All the code in this file is open source
  so feel free to use it any way you would
  like.
  
  design and implementation by
  ArchMiffo, member of Darkbits
----------------------------------------- */

/* -------------------
  Include directives
------------------- */
#include "configparser.hpp"
#include "archio.hpp"
#include <sstream>
#include <queue>
#include <fstream>

// aul stands for ArchMiffo Utility Library
namespace aul
{

/* --------------------------------------
  ParseFile

  This is the main function that parses
  every line in the file
-------------------------------------- */
int ConfigParser::ParseFile(const std::string& filename)
{
  m_Errors.clear();
  bool parseError = false;
  CFGError err;
  int line = 0;
  m_Filename = filename;
  std::ifstream file;
  file.open(m_Filename.c_str(), std::ios::in);

  if(file.fail()) // Couldn't open the file
    return CFG_ERROR::FILE_ERROR;
    
  std::string str;

  while(!file.eof())
  {
    aul::GetLine(file, str, '\n');
    ++line;
    
    if(str.empty())
      continue;
      
    if(str[0] == '#')
        continue;
      
    std::string::size_type index = str.find('=');
    if(index == std::string::npos)
    {
      err.line = line;
      err.code = CFG_ERROR::MISSING_SIGN_ERROR;
      m_Errors.push_back(err);
      parseError = true;
      continue;
    } // end if
    
    index = str.find('#');
    if(index != std::string::npos)
    {
      err.line = line;
      err.code = CFG_ERROR::COMMENT_ERROR;
      m_Errors.push_back(err);
      parseError = true;
      continue;
    } // end if

    std::string command, val;
    std::istringstream is(str);

    aul::GetCommandVal(is, command, val, '=');

    if(Initialize(command, val) != CFG_ERROR::SUCCESS)
    {
      err.line = line;
      err.code = CFG_ERROR::UNKNOWN_COMMAND_ERROR;
      m_Errors.push_back(err);
      parseError = true;
      continue;
    } // end if
  } // end while
  
  if(parseError)
    return CFG_ERROR::PARSE_ERROR;

  return CFG_ERROR::SUCCESS;
} // end ConfigParser::ParseFile

/* --------------------------------------------
  Initialize

  Checks if the command that were found exist
  and if it does it gives the right variable
  its new value
-------------------------------------------- */
int ConfigParser::Initialize(const std::string& command, const std::string& val)
{
  ConfigMap::iterator it = m_ConfigItems.find(command);
  
  if(it == m_ConfigItems.end())
    return CFG_ERROR::UNKNOWN_COMMAND_ERROR;

  std::stringstream in(val);
  (it->second)->Read(in);
  
  return CFG_ERROR::SUCCESS;
} // end ConfigParser::Initialize

/* --------------------------------------------
  SaveFile

  Save the current settings. Doing this will
  not change the format of the config file.
  In other words the comments and everything
  will look exactly as before (except the new
  values of course)
-------------------------------------------- */
int ConfigParser::SaveFile()
{
  std::queue<std::string> tempque;
  std::string str;
  ConfigMap::iterator it;
    
  {
    std::fstream file;
    file.open(m_Filename.c_str(), std::ios::out | std::ios::in);

    if(file.fail()) // Couldn't open the file
      return CFG_ERROR::FILE_ERROR;
  
    while(!file.eof())
    {
      aul::GetLine(file, str, '\n');
      tempque.push(str);
    } // end while
  } // end block
  
  std::fstream file;
  file.open(m_Filename.c_str(), std::ios::out | std::ios::trunc);

  if(file.fail()) // Couldn't open the file
    return CFG_ERROR::FILE_ERROR;

  while(!tempque.empty())
  {
    str = tempque.front();
    tempque.pop();
    
    std::string::size_type index = str.find('=');
    if(index == std::string::npos || str[0] == '#')
    {
      if(tempque.empty())
        file<<str;
      else
        file<<str<<'\n';
      continue;
    } // end if

    std::string::iterator strit = std::find_if(str.begin() + index, str.end(), std::isalnum);

    std::string command, val;
    std::istringstream is(str);

    aul::GetCommandVal(is, command, val, '=');
    
    it = m_ConfigItems.find(command);
    
    if(it == m_ConfigItems.end())
    {
      file<<str<<'\n';
      continue;
    } // end if
    
    std::string temp;
    std::ostringstream os(temp);
    (it->second)->Write(os);

    m_ConfigItems.erase(it);
    
    temp = os.str();
    
    str.erase(strit, str.end());
    str += temp;
    
    file<<str<<'\n';
  } // end while
  
  it = m_ConfigItems.begin();
  for(it; it != m_ConfigItems.end(); ++it)
  {
    file<<it->first<<'=';
    (it->second)->Write(file);
    file<<'\n';
  } // end for

  return CFG_ERROR::SUCCESS;
} // end ConfigParser::SaveFile

/* -----------------------------------------
  PrintErrors
  
  Print all the errors that occured during
  parsing of the config file
----------------------------------------- */
void ConfigParser::PrintErrors(std::ostream& out)
{
  if(m_Errors.empty())
  {
    out<<"No errors found in file "<<m_Filename<<std::endl;
    return;
  } // end if
  
  out<<"The following errors were found in "<<m_Filename<<":"<<std::endl;
  
  for(int i=0; i != m_Errors.size(); ++i)
  {
    out<<"error: "<<m_Errors[i].code<<"   line: "<<m_Errors[i].line<<'\t';
    
    switch(m_Errors[i].code)
    {
      case CFG_ERROR::COMMENT_ERROR :
        out<<"\'#\' sign not at start of line"<<std::endl;
        break;

      case CFG_ERROR::UNKNOWN_COMMAND_ERROR :
        out<<"Command could not be found"<<std::endl;
        break;
        
      case CFG_ERROR::MISSING_SIGN_ERROR :
        out<<"Missing \'=\' sign between the command and value"<<std::endl;
        break;
        
      default : break;
    } // end switch
  } // end for
} // end ConfigParser::PrintErrors

} // end aul