/* -----------------------------------------
  configparser.hpp
  
  All the code in this file is open source
  so feel free to use it any way you would
  like.
  
  design and implementation by 
  ArchMiffo, member of Darkbits
----------------------------------------- */

#ifndef AUL_CONFIGPARSER_HPP
#define AUL_CONFIGPARSER_HPP

/* -------------------
  Include directives
------------------- */
#include <string>
#include <map>
#include <vector>
#include <iostream>
#include "countedptr.hpp"

// aul stands for ArchMiffo Utility Library
namespace aul
{
  
  /* ------------------------
    The CFG_ERROR namespace
  ------------------------ */
  namespace CFG_ERROR
  { 
    // An enum for error codes
    enum
    {
      SUCCESS = 100,
      FILE_ERROR = 101, 
      PARSE_ERROR = 102,
      COMMENT_ERROR = 103,
      UNKNOWN_COMMAND_ERROR = 104,
      MISSING_SIGN_ERROR = 105 
    }; // end enum
  } // end CFG_ERROR

/* --------------------------------
  The CFGError struct
  
  Used to store error information
-------------------------------- */
struct CFGError
{
  int line;
  int code;
}; // end CFGError

/* --------------------------------------------------
  The BasicConfigItem class

  This is a base class that allows us to store any
  type of the template class ConfigItem in a vector
  or some other kind of container class
-------------------------------------------------- */
class BasicConfigItem
{
public :
  virtual ~BasicConfigItem() {}
  virtual std::istream& Read(std::istream& in) = 0;
  virtual std::ostream& Write(std::ostream& out) = 0;
}; // end BasicConfigItem

/* ------------------------------------------------
  The ConfigItem class

  This class is a placeholder for different types 
  of variables. In the ConfigParser class we want
  to store any type to be read from the config
  file. To do this we make use of polymorphism
------------------------------------------------ */
template<typename T>
class ConfigItem : public BasicConfigItem
{
public :
  ConfigItem(T* item, T value = T()) 
  : m_Item(item) { *m_Item = value; }

  std::istream& Read(std::istream& in) { return in>>*m_Item; }
  std::ostream& Write(std::ostream& out) { return out<<*m_Item; }

private :
  T* m_Item;
}; // end ConfigItem

/* ---------------------------------------------
  The ConfigParser class

  Whith this class you can save time by not
  having to write the same code over and over
  every time you want to have a config file.
  Another thing you don't have to do is change
  any config code if you want to add a new
  command in the config file. You simply use
  AddItem and the parse the file
--------------------------------------------- */
class ConfigParser
{
public :
  int ParseFile(const std::string& filename);
  int SaveFile();
  
  template<typename T>
  void AddItem(const std::string& command, T* item, T val=T())
  { m_ConfigItems[command] = 
    aul::counted_ptr<BasicConfigItem>(new ConfigItem<T>(item, val)); }
    
  const std::vector<CFGError>& GetErrors() { return m_Errors; }
  void PrintErrors(std::ostream& out);

private :
  int Initialize(const std::string& command, const std::string& val);

  typedef std::map<std::string, aul::counted_ptr<BasicConfigItem> > ConfigMap;
  ConfigMap m_ConfigItems;
  std::vector<CFGError> m_Errors;
  std::string m_Filename;
}; // end ConfigParser

} // end aul

#endif // end AUL_CONFIGPARSER_HPP